from datetime import datetime
import json
from logging import Logger
import os
from typing import Any
from punkz.core.cache.providers import BaseCacheProvider
from punkz.packages.fs_manager import FileSystemManager


class LocalCacheProvider(BaseCacheProvider):
    def __init__(self, base_directory: str, logger: Logger | None = None) -> None:
        super().__init__()
        self.logger = logger
        self.base_directory = base_directory
        self.fs_manager = FileSystemManager(self.logger)

    def read(
        self, file_name: str, func_name: str, cache_instance_id: str
    ) -> Any | None:
        cache_dir = self._get_cache(cache_instance_id, func_name, self.base_directory)
        file_path = os.path.join(os.getcwd(), cache_dir, file_name)
        if os.path.exists(file_path):
            with open(file_path, "r", encoding="utf-8") as f:
                cached_result = self._parse_types(json.loads(f.read()))
            if cached_result and self._is_cached_result_valid(
                cached_result["metadata"]["expiration"]
            ):
                return cached_result["cached_data"]
            else:
                self.logger.debug("This is expired") if self.logger else None
        return None

    def write(
        self,
        file_name: str,
        func_name: str,
        content: str,
        expiration_date: str,
        cache_instance_id: str,
    ) -> None:
        cache_result = {
            "cached_data": str(content),
            "metadata": {
                "timestamp": str(
                    datetime.now().strftime(self._default_datetime_str_format)
                ),
                "expiration": expiration_date,
            },
        }
        cache_dir = self._get_cache(cache_instance_id, func_name, self.base_directory)
        file_path = os.path.join(os.getcwd(), cache_dir, file_name)
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(cache_result))

    def _get_cache(
        self, cache_instance_id: str, func_name: str, base_dir: str = None
    ) -> str:
        """
        Get cache directory for a function call.
        Parameters:
            func_name (str): name of the function.
        Returns:
            cache_dir (str): cache directory for the function call.
        """
        cache_dir = self._get_cache_dir_name(cache_instance_id, func_name, base_dir)
        self.fs_manager.create_directory(cache_dir)
        return cache_dir

    def _get_cache_dir_name(
        self, cache_instance_id: str, func_name: str, base_dir: str
    ) -> str:
        """
        Create cache directory path for a function call.
        Parameters:
            func_name (str): name of the function.
        Returns:
            cache_dir (str): cache directory for the function call.
        """
        cache_name = func_name + "_cache"
        return os.path.join(self.base_directory, cache_instance_id, cache_name)



import os
from logging import Logger
from typing import Any, Callable, Iterator


class FileSystemManager:
    def __init__(self, logger: Logger | None = None) -> None:
        self.logger = logger

    def create_file(self, file_name: str, content: str) -> None:
        """
        Creates a new file with the given file name and content.

        Args:
            file_name (str): The name of the file to create.
            content (str): The content to write to the file.

        Raises:
            OSError: If an error occurs while creating the file.

        Returns:
            None
        """
        try:
            with open(file_name, "w", encoding="utf-8") as new_file:
                new_file.write(content)
            message = f"File created: {file_name}"
            self.logger.debug(message) if self.logger else print(message)
        except OSError as error:
            message = f"An OSError occurred while creating the file {file_name}"
            self.logger.error(message) if self.logger else print(message)
            self.logger.error(error) if self.logger else print(error)

    def get_filepath_from_directory(self, folder_name: str) -> Iterator[str]:
        """
        Given a folder name, returns an iterator of file paths in that folder.

        Args:
        - folder_name (str): the name of the folder to scan for files

        Returns:
        - Iterator[str]: an iterator of file paths in the folder
        """
        for file in self.scan_directory_for_files(folder_name):
            yield os.path.join(folder_name, file)

    def scan_directory_for_files(self, folder_name: str) -> Iterator[str]:
        """
        Scans the specified directory for files and yields their names.

        Args:
            folder_name (str): The name of the directory to scan.

        Yields:
            str: The name of a file in the directory.
        """
        message = f"Scanning directory {folder_name}"
        self.logger.debug(message) if self.logger else None
        with os.scandir(folder_name) as it:
            for entry in it:
                if entry.is_file():
                    yield entry.name

    def create_directory(self, directory: str) -> None:
        """
        Create a directory if it doesn't exist.

        Args:
            directory (str): The directory to create.

        Returns:
            None
        """
        if not os.path.exists(directory):
            os.makedirs(directory)
            message = f"Folder created: {directory}"
            self.logger.debug(message) if self.logger else None
            return None
        message = f"Folder already exists: {directory}"
        self.logger.debug(message) if self.logger else None

    def get_filepath(self, filepath: str, func: Callable, *args: Any) -> str:
        """
        Given a filepath, checks if the file exists.
        If it does, returns the filepath, otherwise
        returns the result of the function.

        Args:
            filepath (str): The filepath to check.
            func (Callable): The function to call if the file does not exist.

        Returns:
            str: The filepath or the result of the function.
        """
        if os.path.exists(filepath):
            return filepath
        else:
            return func(*args)



import logging
import os


def get_logger(
    logger_name: str, logger_level: str = "INFO", logger_file_level: str = "DEBUG"
) -> logging.Logger:
    """
    This function creates and returns a custom logger.
    By default, it prints infos, warning and errors in the console and saves also the debugs in the .log file

    Args:
        logger_name (string): The name of the Logger
        logger_level (string): The level of logs you want to see in the terminal
        logger_file_level (string): The level of logs you want to have in the .log file

    Returns:
        type: Logger

    Raises:
        AssertionError: If the levels of 'logger_level' and 'logger_file_level' are not
        one of the standard logging levels (NOTSET | INFO | DEBUG | WARNING | ERROR | CRITICAL)

    Example:
        >>> my_logger = get_logpunk("my_module")
        >>> my_logger
        <Logger my_module (DEBUG)>
    """

    # Initialize map
    level_map = {
        "NOTSET": logging.NOTSET,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }

    # Test input parameters
    assert (
        logger_level in level_map.keys()
    ), "[LOG ERROR] 'logger_level' should be: NOTSET | INFO | DEBUG | WARNING | ERROR | CRITICAL"
    assert (
        logger_file_level in level_map.keys()
    ), "[LOG ERROR] 'logger_file_level' should be:  NOTSET | INFO | DEBUG | WARNING | ERROR | CRITICAL"

    # Initialize logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)

    # log_format = "| %(asctime)s | %(levelname)s | source: %(name)s |--> %(message)s"
    # log_format = "| {asctime} | {levelname} | source: {name} |--> {message}"
    log_format = "|{:^13}|{:^17}|{:^17}|--> {}".format(
        "%(asctime)s", "%(levelname)s", "%(name)s", "%(message)s"
    )

    # Create a console handler and set the level to 'logger_level'
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level_map[logger_level])
    console_handler.setFormatter(logging.Formatter(log_format))

    # Define the name and the path of the .log file
    directory = "./.logs"
    if not os.path.exists(directory):
        os.mkdir(directory)
    filename = logger_name.strip().lower().replace(" ", "_") + ".log"
    filepath = os.path.join(directory, filename)

    # Create a file handler and set the level to 'logger_file_level'
    file_handler = logging.FileHandler(filepath)
    file_handler.setLevel(level_map[logger_file_level])
    file_handler.setFormatter(logging.Formatter(log_format))

    # Add the handlers to the logger
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)

    return logger



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Tuple
import traceback


class Processable(ABC):
    """
    Abstract class representing a processable object.
    """

    @abstractmethod
    def invoke(self, data: Any) -> Any:
        """
        Invoke the processable object.
        """
        pass

    @abstractmethod
    def __or__(self, other) -> "Routine":
        """
        Define the pipe operator for the processable object.
        """
        pass


class ProcessStatus(Enum):
    """
    Enum representing the status of a process.

    Attributes:
        GREEN (int): The process is in a green status.
        RED (int): The process is in a red status.
    """

    GREEN = 1
    RED = 0


class ProcessorInput(Processable):
    """
    Represents the input data for a processor.
    """

    def __init__(self, input_data: dict) -> Any:
        self.__data = input_data

    def get_input(self) -> Any:
        """
        Get the input data.
        """
        return self.__data

    def invoke(self, _) -> Any:
        """
        Invoke the processor.
        """
        return {"status": ProcessStatus.GREEN, "data": self.get_input()}

    def __repr__(self) -> str:
        return f"ProcessorInput | data: {self.get_input()})"

    def __or__(self, other) -> "Routine":
        return Routine(self, other)


class Processor(Processable):
    """
    A class representing a processor.

    Attributes:
        __name: The name of the processor.

    Methods:
        get_name: Get the name of the processor.
        process: Process the data.
        invoke: Invoke the processor with data.
        __or__: Create a routine with another processor.
        __repr__: Get a string representation of the processor.
    """

    def __init__(self, name: str) -> None:
        self.__name = name

    def get_name(self):
        return self.__name

    @staticmethod
    def railway(func):
        """
        Decorator function that wraps another function and handles exceptions.

        Args:
            func (function): The function to be wrapped.

        Returns:
            function: The wrapped function.

        Decorator logic:
        This decorator wraps a function with error handling logic.
        It catches any exceptions that occur during the execution
        of the wrapped function and returns a dictionary with the status and data.
        If the status is RED, it returns the original data.
        If the status is GREEN, it calls the wrapped function with
        the provided data and returns the result.
        If an exception occurs, it returns a dictionary with the status set to RED,
        and includes the exception, broken processor name, and input data.

        """

        def wrapper(*args, **kwargs):
            try:
                invoke_kwargs = kwargs["invoke_kwargs"]
                status = invoke_kwargs["status"]
                data = invoke_kwargs["data"]
                self = args[0]
                if status == ProcessStatus.RED:
                    return {"status": status, "data": data}
                return {"status": ProcessStatus.GREEN, "data": func(self, data)}
            except Exception as e:
                return {
                    "status": ProcessStatus.RED,
                    "data": {
                        "exception": e,
                        "broken_processor": self.get_name(),
                        "input_data": kwargs,
                    },
                    "traceback": traceback.format_exc(),
                }

        return wrapper

    @railway
    def process(self, data: dict) -> dict:
        """
        Process the given data and return the modified data.
        This is the method that should be overridden by subclasses.
        Here goes the logic of the processor.

        Args:
            data (dict): The input data to be processed.

        Returns:
            dict: The processed data.

        """
        return data + "_" + self.get_name()

    def invoke(self, data: dict) -> dict:
        """
        Invoke the processor with the given data.

        Args:
            data (dict): The input data for the processor.

        Returns:
            dict: The processed data.

        """
        return self.process(invoke_kwargs=data)

    def __or__(self, other) -> "Routine":
        """
        Performs the logical OR operation between two processors.
        This allows the concatenation of processors in a Routine.

        Args:
            other (Processor): The other processor to perform the OR operation with.

        Returns:
            Routine: A new Routine object representing a pipeline of the two processors.
        """
        return Routine(self, other)

    def __repr__(self) -> str:
        return f"Processor({self.get_name()})"


class Routine(Processable):
    """
    A class representing a pipeline of processors.

    Attributes:
        processors (list): A list of processors in the routine.

    Methods:
        __init__(*processors): Initializes the Routine object with a list of processors.
        __or__(other): Adds a processor to the routine using the '|' operator.
        invoke(input_data=None): Invokes the routine by sequentially applying each processor to the input data.
        __repr__(): Returns a string representation of the Routine object.
    """

    def __init__(self, *processors) -> None:
        self.processors = [*processors]

    def __or__(self, other) -> "Routine":
        self.processors.append(other)
        return self

    def invoke(self, input_data=None) -> Any:
        for processor in self.processors:
            input_data = processor.invoke(input_data)
        return input_data

    def __repr__(self) -> str:
        return f"Routine | processors: {self.processors})"


if __name__ == "__main__":

    class T(Processor):
        def __init__(self, name: str) -> None:
            super().__init__(name)

        @Processor.railway
        def process(self, data) -> Tuple:
            return data["my_data"]

    class E(Processor):
        def __init__(self, name: str) -> None:
            super().__init__(name)

        @Processor.railway
        def process(self, data) -> Tuple:
            return data * self.get_name()

    input_processor = T("T")
    processor_a = Processor("A")
    processor_b = Processor("B")
    processor_c = Processor("C")
    error_processor = E("E")

    input_ = ProcessorInput({"my_data": "my_data"})
    x = input_ | input_processor | processor_a | processor_b | processor_c
    res = x.invoke()
    print(x)
    print("\n------------------\n")
    print(res)



from punkz.core.cache.providers.abstract import BaseCacheProvider
from punkz.core.cache.providers.local_provider import LocalCacheProvider
from punkz.core.cache.providers.aws_s3_provider import AWSS3CacheProvider



import os
import sys
import threading


class ProgressPercentage(object):
    def __init__(self, filename: str, filesize: float, operation_type: str) -> None:
        self.operation_type = operation_type
        self._filename = filename
        self._size = filesize
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount: int) -> None:
        """
        Update the progress bar with the number of bytes received.

        Args:
            bytes_amount: The number of bytes received.
        """
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write(
                "\r%s %s  %s / %s  (%.2f%%)"
                % (
                    self.operation_type,
                    self._filename,
                    self._seen_so_far,
                    self._size,
                    percentage,
                )
            )
            sys.stdout.flush()



from logging import Logger
from typing import Any
from punkz.packages.aws.s3 import S3
from punkz.packages.log import get_logger
import boto3


class AWS:
    def __init__(
        self, logger: Logger | None, access_key_id: str, secret_access_key: str
    ) -> None:
        self.logger = logger

        self.aws_access_key_id = access_key_id
        self.aws_secret_access_key = secret_access_key
        boto3.setup_default_session(
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )
        self.S3 = S3(self.aws_access_key_id, self.aws_secret_access_key, logger)



from io import BytesIO
from logging import Logger
from typing import Any
import boto3
import botocore.exceptions
from punkz.packages.aws.progress import ProgressPercentage
import os


class S3:
    def __init__(
        self, aws_access_key_id: str, aws_secret_access_key: str, logger: Logger | None
    ) -> None:
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.logger = logger
        self.s3 = boto3.resource("s3")

    def get_file_list_from_s3(self, bucket_name: str) -> Any:
        """Get a list of files from an S3 bucket
        Args:
            bucket_name: Bucket to list files from
        Returns:
            List of files in the bucket
        """
        bucket = self.s3.Bucket(bucket_name)
        list_of_files = []
        for my_bucket_object in bucket.objects.all():
            list_of_files.append(my_bucket_object.key)
        message = f"List of files correctly retrieved from {bucket_name}"
        self.logger.debug(message) if self.logger else None
        return list_of_files

    def download_file(
        self,
        bucket_name: str,
        object_key: str,
        local_file_path: str,
        verbose: bool = False,
    ) -> Any:
        """Download a file to an S3 bucket
        Args:
            bucket_name: Bucket to download from
            object_key: S3 object key
            local_file_path: Path to the file you want to save (e.g. my_data/my_file.txt)
        Returns:
            The file from AWS S3
        """

        try:
            callback = (
                None
                if not verbose
                else ProgressPercentage(
                    filename=local_file_path,
                    filesize=self.s3.meta.client.head_object(
                        Bucket=bucket_name, Key=object_key
                    )["ContentLength"],
                    operation_type="Downloading",
                ),
            )

            self.s3.meta.client.download_file(
                bucket_name,
                object_key,
                local_file_path,
                Callback=callback,
            )

            message = f"File downloaded successfully: {local_file_path}"
            self.logger.debug(message) if self.logger else None
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                message = f"The object '{object_key}' does not exist in the bucket '{bucket_name}'."
                self.logger.error(message) if self.logger else print(message)
            else:
                message = f"Error downloading file: {e}"
                self.logger.error(message) if self.logger else print(message)

    def download_bytes(
        self, bucket_name: str, object_key: str, verbose: bool = False
    ) -> BytesIO:
        """Download a file to an S3 bucket as bytes
        Args:
            bucket_name: Bucket to download from
            object_key: S3 object key
            local_file_path: Path to the file you want to save (e.g. my_data/my_file.txt)
        Returns:
            The file from AWS S3
        """
        try:
            callback = (
                None
                if not verbose
                else ProgressPercentage(
                    filename=object_key,
                    filesize=self.s3.meta.client.head_object(
                        Bucket=bucket_name, Key=object_key
                    )["ContentLength"],
                    operation_type="Downloading",
                )
            )

            bytes_file = BytesIO()
            self.s3.meta.client.download_fileobj(
                bucket_name,
                object_key,
                bytes_file,
                Callback=callback,
            )
            message = f"File downloaded successfully: {object_key}"
            self.logger.debug(message) if self.logger else None
            bytes_file.seek(0)
            return bytes_file
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                message = f"The object '{object_key}' does not exist in the bucket '{bucket_name}'."
                self.logger.error(message) if self.logger else print(message)
            else:
                message = f"Error downloading file: {e}"
                self.logger.error(message) if self.logger else print(message)

    def upload_file(
        self,
        file_name: str,
        bucket_name: str,
        object_name: str = None,
        metadata: dict = None,
        verbose: bool = False,
    ) -> bool:
        """Upload a file to an S3 bucket
        Args:
            file_name: File to upload
            bucket: Bucket to upload to
            object_name: S3 object name. If not specified, file_name is used
            metadata: Metadata to associate with the uploaded object
        Returns:
            True if the file was uploaded, else False
        """
        # If S3 object_name was not specified, use file_name
        if object_name is None:
            object_name = os.path.basename(file_name)

        try:
            extra_args = {}
            if metadata:
                extra_args["Metadata"] = metadata

            file_size = float(os.path.getsize(file_name))
            callback = (
                None
                if not verbose
                else ProgressPercentage(
                    filename=file_name,
                    filesize=file_size,
                    operation_type="Uploading",
                )
            )

            response = self.s3.meta.client.upload_file(
                file_name,
                bucket_name,
                object_name,
                Callback=callback,
                ExtraArgs=extra_args,
            )
            message = f"File uploaded successfully: {file_name}"
            self.logger.debug(message) if self.logger else None
        except botocore.exceptions.ClientError as e:
            message = f"Error uploading file: {e}"
            self.logger.error(message) if self.logger else print(message)
            return False
        return True

    def upload_bytes(
        self,
        file: BytesIO,
        bucket_name: str,
        object_name: str,
        metadata: dict = None,
        verbose: bool = False,
    ) -> bool:
        """Upload a stream of bytes to an S3 bucket
        Args:
            file: File to upload in BytesIO format
            bucket: Bucket to upload to
            object_name: S3 object name
            metadata: Metadata to associate with the uploaded object
        Returns:
            True if the file was uploaded, else False
        """
        try:
            extra_args = {}
            if metadata:
                extra_args["Metadata"] = metadata

            file_size = float(len(file.getbuffer()))

            callback = (
                None
                if not verbose
                else ProgressPercentage(
                    filename=object_name,
                    filesize=file_size,
                    operation_type="Uploading",
                )
            )

            response = self.s3.meta.client.upload_fileobj(
                file,
                bucket_name,
                object_name,
                Callback=callback,
                ExtraArgs=extra_args,
            )
            message = f"Object uploaded successfully: {object_name}"
            self.logger.debug(message) if self.logger else None
        except botocore.exceptions.ClientError as e:
            message = f"Error uploading object: {e}"
            self.logger.error(message) if self.logger else print(message)
            return False
        return True

    def get_object_metadata(self, bucket_name: str, object_key: str) -> dict | None:
        """Get metadata of an object from an S3 bucket
        Args:
            bucket_name: Bucket containing the object
            object_key: S3 object key
        Returns:
            Metadata of the S3 object as a dictionary
        """
        try:
            response = self.s3.meta.client.head_object(
                Bucket=bucket_name, Key=object_key
            )
            message = f"Metadata of object '{object_key}' retrieved successfully."
            self.logger.debug(message) if self.logger else None
            return response.get("Metadata", {})
        except botocore.exceptions.ClientError as e:
            message = f"Error getting object metadata: {e}"
            self.logger.debug(message) if self.logger else None
            return None



from abc import ABC, abstractmethod
from ast import literal_eval
from datetime import datetime
from typing import Any


class BaseCacheProvider(ABC):
    def __init__(self) -> None:
        self._default_datetime_str_format = "%d-%m-%Y %H:%M:%S"

    @abstractmethod
    def read(self, file_name: str, func_name: str) -> Any:
        pass

    @abstractmethod
    def write(self, file_name: str, func_name: str, content: str) -> None:
        pass

    def _is_cached_result_valid(self, expiration_date: str) -> bool:
        """
        Check if the cached result is valid.
        Parameters:
            expiration_date (str): expiration date of the cached result.
        Returns:
            True if the cached result is valid, False otherwise.
        """
        if expiration_date == "never":
            return True
        else:
            return datetime.now() < datetime.strptime(
                expiration_date, self._default_datetime_str_format
            )

    def _has_expiration_changed(
        self, expiration_date: str, cached_expiration_date
    ) -> bool:
        """
        Check if the expiration date has changed.
        Parameters:
            expiration_date (str): expiration date of the cached result.
        Returns:
            True if the expiration date has changed, False otherwise.
        """
        return cached_expiration_date != expiration_date

    @staticmethod
    def _parse_types(res: str) -> Any:
        """
        Parse the type of the result.
        Parameters:
            res (str): result of the function call.
        Returns:
            The parsed result.
        """
        try:
            parsed_res = literal_eval(res)
        except:
            parsed_res = res

        return parsed_res



from ast import literal_eval
from datetime import datetime
import hashlib
from logging import Logger
import os
from typing import Any, Callable
import json
from datetime import datetime, timedelta
from punkz.core.cache import LocalCacheProvider
from punkz.core.cache import AWSS3CacheProvider


class Cache:
    def __init__(
        self,
        provider: LocalCacheProvider | AWSS3CacheProvider,
        cache_instance_id: str,
        expiration: str | None = None,
        logger: Logger | None = None,
    ) -> None:
        self.logger = logger
        self._default_datetime_str_format = "%d-%m-%Y %H:%M:%S"
        self._deltatime_unit_mapping = {
            "d": "days",
            "h": "hours",
            "m": "minutes",
            "s": "seconds",
        }
        self.provider = provider
        self.expiration_date = self._compute_expiration_date(expiration)
        self.cache_instance_id = cache_instance_id

    def _compute_expiration_date(self, delta_str: str | None) -> str:
        """
        Compute the expiration date based on the given delta string.

        Args:
            delta_str (str | None): The delta string representing the time duration.

        Returns:
            str: The expiration date in the format specified by self._default_datetime_str_format.

        Raises:
            ValueError: If the time unit is invalid.

        """
        if delta_str:
            # Parse the input string to extract the value and unit
            value, unit = int(delta_str[:-1]), delta_str[-1]
            # Validate the unit and calculate the timedelta
            if unit in self._deltatime_unit_mapping:
                delta = timedelta(**{self._deltatime_unit_mapping[unit]: value})
                future_date = datetime.now() + delta
                return future_date.strftime(self._default_datetime_str_format)
            else:
                raise ValueError(
                    'Invalid time unit. Use "d" for days, "h" for hours. E.g. "45d" or "2h"'
                )
        return "never"

    def _sha256_hash(self, string: str) -> str:
        """
        Get sha256 hash of a string.
        Parameters:
            string (str): string to be hashed.
        Returns:
            hash (str): sha256 hash of the string.
        """
        hash = hashlib.sha256()
        hash.update(string.encode("utf-8"))
        return hash.hexdigest()

    def _normalize_text(self, text: str) -> str:
        """
        Normalize the given text by converting it to lowercase, removing leading and trailing whitespaces,
        and replacing multiple whitespaces with a single space.

        Args:
            text (str): The text to be normalized.

        Returns:
            str: The normalized text.
        """
        normalized_text = text.lower()

        # Remove leading and trailing whitespaces
        normalized_text = normalized_text.strip()

        # Replace multiple whitespaces with a single space
        normalized_text = " ".join(normalized_text.split())

        return normalized_text

    def _build_filename_from_params(self, *args, **kwargs) -> str:
        """
        Get hash key for a function call.
        Parameters:
            *args (list): list of arguments.
            **kwargs (dict): dictionary of keyword arguments.
        Returns:
            hash_key (str): hash key for the function call.
        """
        name: str = ""
        for arg in args:
            name += f"_{str(arg)}"
        for key, value in kwargs.items():
            name += f"_{str(key)}" + f"_{str(value)}"
        name = self._normalize_text(name).replace(" ", "_")
        return name

    def cache(self, func) -> Callable:
        """
        Decorator to cache the results of a function call.
        Parameters:
            func (function): function to be cached.
            base_dir (str): directory to store the cached results.
            use_sha256 (bool): whether to use sha256 hash for the cached file name.
        Returns:
        """
        self.logger.debug(f"Cache decorator called for function {func.__name__}")
        def wrapper(*args, **kwargs):
            hash_key = (
                func.__name__
                + "_"
                + self._sha256_hash(self._build_filename_from_params(*args, **kwargs))
            )
            file_name = hash_key + ".json"
            self.logger.debug(f"Cache file name: {file_name}")
            result = self.provider.read(
                file_name, func.__name__, self.cache_instance_id
            )
            self.logger.debug(f"Cache result: {result}")
            if not result:
                self.logger.debug("Cache miss")
                result = func(*args, **kwargs)
                self.provider.write(
                    file_name,
                    func.__name__,
                    result,
                    self.expiration_date,
                    self.cache_instance_id,
                )

            return result

        return wrapper



from datetime import datetime
from io import BytesIO
import json
from logging import Logger
import os
from typing import Any
from punkz.core.cache.providers import BaseCacheProvider
from punkz.packages.aws import AWS


class AWSS3CacheProvider(BaseCacheProvider):
    def __init__(
        self,
        bucket_name: str,
        access_key_id: str,
        secret_access_key: str,
        logger: Logger | None = None,
    ) -> None:
        super().__init__()
        self.logger = logger
        self.bucket_name = bucket_name
        self.aws = AWS(self.logger, access_key_id, secret_access_key)

    def read(
        self, file_name: str, func_name: str, cache_instance_id: str
    ) -> Any | None:
        cache_subfolder_name = func_name + "_cache"
        object_key = os.path.join(cache_instance_id, cache_subfolder_name, file_name)
        s3_object_metadata = self.aws.S3.get_object_metadata(
            self.bucket_name, object_key
        )
        if s3_object_metadata and self._is_cached_result_valid(
            s3_object_metadata["expiration"]
        ):
            bytes_content = self.aws.S3.download_bytes(
                self.bucket_name,
                object_key,
            )
            content_as_bytes = bytes_content.read()
            content = self._parse_types(content_as_bytes.decode("utf-8"))
            if content:
                return content["cached_data"]
        return None

    def write(
        self,
        file_name: str,
        func_name: str,
        content: str,
        expiration_date: str,
        cache_instance_id: str,
    ) -> None:
        cache_result = {
            "cached_data": str(content),
            "metadata": {
                "timestamp": str(
                    datetime.now().strftime(self._default_datetime_str_format)
                ),
                "expiration": expiration_date,
            },
        }
        metadata = {
            "expiration": expiration_date,
        }
        json_cache_result = json.dumps(cache_result)
        in_memory_file = BytesIO(json_cache_result.encode("utf-8"))
        cache_subfolder_name = func_name + "_cache"
        object_key = os.path.join(cache_instance_id, cache_subfolder_name, file_name)
        self.aws.S3.upload_bytes(in_memory_file, self.bucket_name, object_key, metadata)

